from rita.utils import ExtendedOp


def ORTH(value, config, op=None):
    """
    Ignores case-insensitive configuration and checks words as written
    that means case-sensitive even if configuration is case-insensitive
    """
    new_op = ExtendedOp(op)
    new_op.case_sensitive_override = True
    return "orth", value, new_op



from rita.utils import ExtendedOp


def REGEX(regex_pattern, config, op=None):
    """
    Matches words based on a Regex pattern
    e.g. all words that start with an 'a' would be
    REGEX("^a")
    """
    new_op = ExtendedOp(op)
    new_op.local_regex_override = True
    return "regex", regex_pattern, new_op



import logging

from functools import partial
from typing import Any, TYPE_CHECKING, Mapping, Callable, Generator, AnyStr

from rita.utils import ExtendedOp
from rita.types import Rules, Patterns

logger = logging.getLogger(__name__)

SpacyPattern = Generator[Mapping[AnyStr, Any], None, None]
ParseFn = Callable[[Any, "SessionConfig", ExtendedOp], SpacyPattern]

if TYPE_CHECKING:
    # We cannot simply import SessionConfig because of cyclic imports
    from rita.config import SessionConfig


def any_of_parse(lst, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    if op.ignore_case(config):
        normalized = sorted([item.lower()
                             for item in lst])
        base = {"LOWER": {"REGEX": r"^({0})$".format("|".join(normalized))}}
    else:
        base = {"TEXT": {"REGEX": r"^({0})$".format("|".join(sorted(lst)))}}

    if not op.empty():
        base["OP"] = op.value
    yield base


def regex_parse(r, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    if op.ignore_case(config):
        d = {"LOWER": {"REGEX": r.lower()}}
    else:
        d = {"TEXT": {"REGEX": r}}

    if not op.empty():
        d["OP"] = op.value
    yield d


def fuzzy_parse(r, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    # TODO: build premutations
    d = {"LOWER": {"REGEX": "({0})[.,?;!]?".format("|".join(r))}}
    if not op.empty():
        d["OP"] = op.value
    yield d


def generic_parse(tag, value, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    d = {}
    if tag == "ORTH" and op.ignore_case(config):
        d["LOWER"] = value.lower()
    else:
        d[tag] = value

    if not op.empty():
        d["OP"] = op.value
    yield d


def entity_parse(value, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    tag = "ENT_TYPE"
    if op.empty():
        op.op = "+"
    return generic_parse(tag, value, config, op)


def punct_parse(_, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    d = dict()
    d["IS_PUNCT"] = True
    if not op.empty():
        d["OP"] = op.value
    yield d


def any_parse(_, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    d = dict()
    if not op.empty():
        d["OP"] = op.value
    yield d


def phrase_parse(value, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    """
    TODO: Does not support operators
    """
    splitter = next((s for s in ["-", " "]
                     if s in value), None)
    if splitter:
        buff = value.split(splitter)
        yield next(generic_parse("ORTH", buff[0], config=config, op=ExtendedOp()))
        for b in buff[1:]:
            if splitter != " ":
                yield next(generic_parse("ORTH", splitter, config=config, op=ExtendedOp()))
            yield next(generic_parse("ORTH", b, config=config, op=ExtendedOp()))
    else:
        yield next(generic_parse("ORTH", value, config=config, op=ExtendedOp()))


def tag_parse(values, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    """
    For generating POS/TAG patterns based on a Regex
    e.g. TAG("^NN|^JJ") for adjectives or nouns
    also deals with TAG_WORD for tag and word or tag and list
    """
    d = {"TAG": {"REGEX": values["tag"]}}
    if "word" in values:
        if op.ignore_case(config):
            d["LOWER"] = values["word"].lower()
        else:
            d["TEXT"] = values["word"]
    elif "list" in values:
        lst = values["list"]
        if op.ignore_case(config):
            normalized = sorted([item.lower()
                                 for item in lst])
            d["LOWER"] = {"REGEX": r"^({0})$".format("|".join(normalized))}
        else:
            d["TEXT"] = {"REGEX": r"^({0})$".format("|".join(sorted(lst)))}
    if not op.empty():
        d["OP"] = op.value
    yield d


def nested_parse(values, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    from rita.macros import resolve_value
    results = rules_to_patterns("", [resolve_value(v, config=config)
                                     for v in values], config=config)
    return results["pattern"]


def orth_parse(value, config: "SessionConfig", op: ExtendedOp) -> SpacyPattern:
    d = {}
    d["ORTH"] = value
    if not op.empty():
        d["OP"] = op.value
    yield d


PARSERS: Mapping[str, ParseFn] = {
    "any_of": any_of_parse,
    "any": any_parse,
    "value": partial(generic_parse, "ORTH"),
    "regex": regex_parse,
    "entity": entity_parse,
    "lemma": partial(generic_parse, "LEMMA"),
    "pos": partial(generic_parse, "POS"),
    "punct": punct_parse,
    "fuzzy": fuzzy_parse,
    "phrase": phrase_parse,
    "tag": tag_parse,
    "nested": nested_parse,
    "orth": orth_parse,
}


def rules_to_patterns(label: str, data: Patterns, config: "SessionConfig"):
    logger.debug(data)
    return {
        "label": label,
        "pattern": [p
                    for (t, d, op) in data
                    for p in PARSERS[t](d, config, ExtendedOp(op))],
    }


def compile_rules(rules: Rules, config: "SessionConfig", **kwargs):
    logger.info("Using spaCy rules implementation")
    return [rules_to_patterns(label, patterns, config=config)
            for (label, patterns) in rules]



from typing import Any, Mapping, Tuple, List, AnyStr

opts = Mapping[Any, Any]
RuleData = Tuple[AnyStr, List[Any], AnyStr]
Patterns = List[RuleData]
RuleGroup = Tuple[AnyStr, Patterns]
Rules = List[RuleGroup]



import logging

from rita.macros import resolve_value
from rita.utils import flatten, ExtendedOp

logger = logging.getLogger(__name__)


STOP_NAMES = {"von", "van", "de", "dos"}


def trim_name(name):
    if name in STOP_NAMES:
        return name
    return name[0] + r"\."


def trim_seniority(name):
    if name.lower() == "junior":
        return r"jr\."
    elif name.lower() == "senior":
        return r"sr\."
    else:
        return name


def remove_empty(x):
    return x.strip() != ""


def generate_names(initial_list):
    """"
    Generates variations of names
    Eg. {First Middle Last; First M. Last; F. M. Last}
    """
    for name in initial_list:
        yield name.strip(),

        buff = name.strip().split(" ")
        if len(buff) == 2:
            yield trim_name(buff[0]), buff[1]
        elif len(buff) == 3:
            if buff[2].lower() == "junior" or buff[2].lower() == "senior":
                yield buff[0], buff[1], trim_seniority(buff[2])
            else:
                yield buff[0], trim_name(buff[1]), buff[2]
                yield trim_name(buff[0]), trim_name(buff[1]), buff[2]


def NAMES(*args, config, op=None):
    if type(args[0]) == list:
        initial_list = [resolve_value(arg, config=config)
                        for arg in flatten(args)]
    else:
        initial_list = [args[0]]

    names = list([" ".join(filter(remove_empty, names))
                  for names in generate_names(initial_list)])
    logger.debug("Generated list of names: {}".format(names))
    new_op = ExtendedOp(op)
    new_op.case_sensitive_override = True
    return "any_of", names, new_op



from rita.utils import ExtendedOp


def TAG(tag, config, op=None):
    """
    For generating POS/TAG patterns based on a Regex
    e.g. TAG("^NN|^JJ") for nouns or adjectives
    """
    values = {"tag": tag}
    return "tag", values, ExtendedOp(op)


def TAG_WORD(tag, value, config, op=None):
    """
    For generating TAG patterns with a word or a list
    e.g. match only "proposed" when it is in the sentence a verb (and not an adjective):
    TAG_WORD("^VB", "proposed")
    e.g. match a list of words only to verbs
    words = {"percived", "proposed"}
    {TAG_WORD("^VB", words)?}->MARK("LABEL")
    """
    values = {"tag": tag}
    if type(value) == list:
        values["list"] = value
    else:
        values["word"] = value
    return "tag", values, ExtendedOp(op)



import json
import logging

from platform import system

from ctypes import (c_char_p, c_int, c_uint, c_long, Structure, cdll, POINTER)
from typing import Any, TYPE_CHECKING, Tuple, List, AnyStr

from rita.engine.translate_standalone import rules_to_patterns, RuleExecutor
from rita.types import Rules

logger = logging.getLogger(__name__)

field = Tuple[AnyStr, Any]
fields = List[field]

if TYPE_CHECKING:
    # We cannot simply import SessionConfig because of cyclic imports
    from rita.config import SessionConfig


class NamedRangeResult(Structure):
    _fields_ = [
        ("start", c_long),
        ("end", c_long),
        ("name", c_char_p),
    ]


class ResultEntity(Structure):
    _fields_ = [
        ("label", c_char_p),
        ("start", c_long),
        ("end", c_long),
        ("sub_count", c_uint),
    ]


class Result(Structure):
    _fields_ = [
        ("count", c_uint)
    ]


class Context(Structure):
    _fields_: fields = []


def load_lib():
    try:
        os_name = system()
        if os_name == "Windows":
            lib = cdll.LoadLibrary("rita_rust.dll")
        elif os_name == "Darwin":
            lib = cdll.LoadLibrary("librita_rust.dylib")
        else:
            lib = cdll.LoadLibrary("librita_rust.so")
        lib.compile.restype = POINTER(Context)
        lib.execute.argtypes = [POINTER(Context), c_char_p]
        lib.execute.restype = POINTER(Result)
        lib.clean_env.argtypes = [POINTER(Context)]
        lib.clean_result.argtypes = [POINTER(Result)]
        lib.read_result.argtypes = [POINTER(Result), c_int]
        lib.read_result.restype = POINTER(ResultEntity)
        lib.read_submatch.argtypes = [POINTER(ResultEntity), c_int]
        lib.read_submatch.restype = POINTER(NamedRangeResult)
        return lib
    except Exception as ex:
        logger.error("Failed to load rita-rust library, reason: {}\n\n"
                     "Most likely you don't have required shared library to use it".format(ex))


class RustRuleExecutor(RuleExecutor):
    def __init__(self, patterns, config: "SessionConfig"):
        self.config = config
        self.context = None

        self.lib = load_lib()
        self.patterns = [self._build_regex_str(label, rules)
                         for label, rules in patterns]

        self.compile()

    @staticmethod
    def _build_regex_str(label, rules):
        indexed_rules = ["(?P<s{}>{})".format(i, r) if not r.startswith("(?P<") else r
                         for i, r in enumerate(rules)]
        return r"(?P<{0}>{1})".format(label, "".join(indexed_rules))

    def compile(self):
        flag = 0 if self.config.ignore_case else 1
        c_array = (c_char_p * len(self.patterns))(*list([p.encode("UTF-8") for p in self.patterns]))
        self.context = self.lib.compile(c_array, len(c_array), flag)
        return self.context

    def execute(self, text, include_submatches=True):
        result_ptr = self.lib.execute(self.context, text.encode("UTF-8"))
        count = result_ptr[0].count
        for i in range(0, count):
            match_ptr = self.lib.read_result(result_ptr, i)
            match = match_ptr[0]
            matched_text = text[match.start:match.end].strip()

            def parse_subs():
                k = match.sub_count
                for j in range(0, k):
                    s = self.lib.read_submatch(match_ptr, j)[0]
                    start = s.start
                    end = s.end
                    sub_text = text[start:end]

                    if sub_text.strip() == "":
                        continue

                    yield {
                        "text": sub_text.strip(),
                        "start": start,
                        "end": end,
                        "key": s.name.decode("UTF-8"),
                    }

            yield {
                "start": match.start,
                "end": match.end,
                "text": matched_text,
                "label": match.label.decode("UTF-8"),
                "submatches": list(parse_subs()) if include_submatches else []
            }

    def clean_context(self):
        self.lib.clean_env(self.context)

    @staticmethod
    def load(path):
        from rita.config import SessionConfig
        config = SessionConfig()
        with open(path, "r") as f:
            patterns = [(obj["label"], obj["rules"])
                        for obj in map(json.loads, f.readlines())]
            return RustRuleExecutor(patterns, config)


def compile_rules(rules: Rules, config: "SessionConfig", **kwargs) -> RustRuleExecutor:
    logger.info("Using rita-rust rule implementation")
    patterns = [rules_to_patterns(*group, config=config) for group in rules]
    executor = RustRuleExecutor(patterns, config)
    return executor



import rita


def setup_spacy(model, patterns=None, rules_path=None, rules_string=None, override_ents=True):
    import spacy
    major, _, _ = spacy.__version__.split(".")
    if major == "2":
        return _spacy_v2(model, patterns, rules_path, rules_string, override_ents)
    elif major == "3":
        return _spacy_v3(model, patterns, rules_path, rules_string, override_ents)
    else:
        raise RuntimeError("Unsupported spaCy version: {}".format(major))


def _spacy_v2(model, patterns=None, rules_path=None, rules_string=None, override_ents=True):
    from spacy.pipeline import EntityRuler
    ruler = EntityRuler(model, overwrite_ents=override_ents)
    if not patterns:
        if rules_path:
            patterns = rita.compile(rules_path, use_engine="spacy")
        elif rules_string:
            patterns = rita.compile_string(rules_string, use_engine="spacy")
        else:
            raise RuntimeError("Please provides rules. Either `patterns`, `rules_path` or `rules_string`")

        ruler.add_patterns(patterns)
    else:
        ruler.from_disk(patterns)

    model.add_pipe(ruler)
    return model


def _spacy_v3(model, patterns=None, rules_path=None, rules_string=None, override_ents=True):
    ruler = model.add_pipe("entity_ruler", config={"overwrite_ents": override_ents, "validate": True})
    if not patterns:
        if rules_path:
            patterns = rita.compile(rules_path, use_engine="spacy")
        elif rules_string:
            patterns = rita.compile_string(rules_string, use_engine="spacy")
        else:
            raise RuntimeError("Please provides rules. Either `patterns`, `rules_path` or `rules_string`")

        ruler.add_patterns(patterns)
    else:
        ruler.from_disk(patterns)
    return model
